#include <Arduino.h>
#include <TM1637Display.h>
#include <Ticker.h>
#ifdef CAPTIVE
#include <DNSServer.h>
#endif
#ifdef ESP32
#include <WiFi.h>
#include <AsyncTCP.h>
#elif defined(ESP8266)
#include <ESP8266WiFi.h>
#include <ESPAsyncTCP.h>
#endif
#include <ESPAsyncWebServer.h>
#include <LittleFS.h>
// local includes
#include "defaults.h"

// Module connection pins (ESP GPIO-Nums)
#define CLK 16
#define DIO_FAC_1_7SEG 14
#define DIO_FAC_2_7SEG 12
#define DIO_FAC_3_7SEG 13

#define DIO_FAC_1_TRG 0
#define DIO_FAC_2_TRG 2
#define DIO_FAC_3_TRG 15

enum Factions
{
  NONE,
  FACTION_1,
  FACTION_2,
  FACTION_3
};

void SevenSeg_Output();
void FactionTicker_callback();
void serialOutTicker_callback();
void inputGetterTicker_callback();

TM1637Display disp_FAC_1(CLK, DIO_FAC_1_7SEG);
TM1637Display disp_FAC_2(CLK, DIO_FAC_2_7SEG);
TM1637Display disp_FAC_3(CLK, DIO_FAC_3_7SEG);

WiFiEventHandler stationConnectedHandler;
WiFiEventHandler stationDisconnectedHandler;

#ifdef CAPTIVE
DNSServer dnsServer;
#endif
AsyncWebServer server(80);

Ticker FactionTicker(FactionTicker_callback, 500, 0, MILLIS);
Ticker SerialOutputTicker(serialOutTicker_callback, 5000, 0, MILLIS);
Ticker InputGetterTicker(inputGetterTicker_callback, 500, 0, MILLIS);

Factions activeFaction = NONE;

uint32_t Count_Faction_1 = 0;
uint32_t Count_Faction_2 = 0;
uint32_t Count_Faction_3 = 0;
uint8_t Faction_1_dot = 0;
uint8_t Faction_2_dot = 0;
uint8_t Faction_3_dot = 0;

String processor(const String &var)
{
  char buffer[16] = {0};

  if (var == "POINTS_FAC_1")
    itoa(Count_Faction_1, buffer, 10);

  if (var == "POINTS_FAC_2")
    itoa(Count_Faction_2, buffer, 10);

  if (var == "POINTS_FAC_3")
    itoa(Count_Faction_3, buffer, 10);

  if (var == "STATUS_FAC_1")
    return activeFaction == FACTION_1 ? "ACTIVE" : "INACTIVE";

  if (var == "STATUS_FAC_2")
    return activeFaction == FACTION_2 ? "ACTIVE" : "INACTIVE";

  if (var == "STATUS_FAC_3")
    return activeFaction == FACTION_3 ? "ACTIVE" : "INACTIVE";

  if (var == "NAME_FAC_1")
    return FACTION_1_NAME;

  if (var == "NAME_FAC_2")
    return FACTION_2_NAME;

  if (var == "NAME_FAC_3")
    return FACTION_3_NAME;

  if (var == "TITLE")
    return WIFI_SSID;

  return String(buffer);
}

String macToString(const unsigned char *mac)
{
  char buf[20];
  snprintf(buf, sizeof(buf), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
  return String(buf);
}

void onStationConnected(const WiFiEventSoftAPModeStationConnected &evt)
{
  Serial.print("Station connected: ");
  Serial.println(macToString(evt.mac));
}

void onStationDisconnected(const WiFiEventSoftAPModeStationDisconnected &evt)
{
  Serial.print("Station disconnected: ");
  Serial.println(macToString(evt.mac));
}

String getValue(String data, char separator, int index)
{
  int found = 0;
  int strIndex[] = {0, -1};
  int maxIndex = data.length() - 1;

  for (int i = 0; i <= maxIndex && found <= index; i++)
  {
    if (data.charAt(i) == separator || i == maxIndex)
    {
      found++;
      strIndex[0] = strIndex[1] + 1;
      strIndex[1] = (i == maxIndex) ? i + 1 : i;
    }
  }

  return found > index ? data.substring(strIndex[0], strIndex[1]) : "";
}

class CaptiveRequestHandler : public AsyncWebHandler
{
public:
  CaptiveRequestHandler() {}
  virtual ~CaptiveRequestHandler() {}

  bool canHandle(AsyncWebServerRequest *request)
  {
    return true;
  }

  void handleRequest(AsyncWebServerRequest *request)
  {

    Serial.print("Requested URL: ");
    Serial.println(request->url());

    //List all parameters
    int params = request->params();
    for (int i = 0; i < params; i++)
    {
      AsyncWebParameter *p = request->getParam(i);
      if (p->isFile())
      { //p->isPost() is also true
        Serial.printf("FILE[%s]: %s, size: %u\n", p->name().c_str(), p->value().c_str(), p->size());
      }
      else if (p->isPost())
      {
        Serial.printf("POST[%s]: %s\n", p->name().c_str(), p->value().c_str());
      }
      else
      {
        Serial.printf("GET[%s]: %s\n", p->name().c_str(), p->value().c_str());
      }
    }

    if (LittleFS.exists(request->url()) == false)
    {
      request->send(404, "text/plain", "Not found");
      return;
    }

    if (request->url() == "/favicon.ico")
    {
      request->send(LittleFS, request->url(), "image/x-icon");
      return;
    }

    if (request->url() == "/" || request->url() == "/index.html")
    {
      request->send(LittleFS, "/index.html", String(), false, processor);
      return;
    }

    if (getValue(request->url(), '/', 1) == "css")
    {
      request->send(LittleFS, request->url(), "text/css");
      return;
    }
  }
};

void SevenSeg_Output()
{
  disp_FAC_1.setBrightness(activeFaction == FACTION_1 ? 7 : 0);
  disp_FAC_2.setBrightness(activeFaction == FACTION_2 ? 7 : 0);
  disp_FAC_3.setBrightness(activeFaction == FACTION_3 ? 7 : 0);

  disp_FAC_1.showNumberDecEx(Count_Faction_1 / 20, Faction_1_dot, true, 4, 0);
  disp_FAC_2.showNumberDecEx(Count_Faction_2 / 20, Faction_2_dot, true, 4, 0);
  disp_FAC_3.showNumberDecEx(Count_Faction_3 / 20, Faction_3_dot, true, 4, 0);
}

void FactionTicker_callback()
{
  switch (activeFaction)
  {
  case FACTION_1:
    Count_Faction_1++;
    Faction_1_dot = Faction_1_dot == 0x80 || Faction_1_dot == 0x00 ? 0x10 : Faction_1_dot << 1;
    Faction_2_dot = 0;
    Faction_3_dot = 0;
    break;

  case FACTION_2:
    Count_Faction_2++;
    Faction_2_dot = Faction_2_dot == 0x80 || Faction_2_dot == 0x00 ? 0x10 : Faction_2_dot << 1;
    Faction_1_dot = 0;
    Faction_3_dot = 0;
    break;

  case FACTION_3:
    Count_Faction_3++;
    Faction_3_dot = Faction_3_dot == 0x80 || Faction_3_dot == 0x00 ? 0x10 : Faction_3_dot << 1;
    Faction_1_dot = 0;
    Faction_2_dot = 0;
    break;

  default:
    break;
  }
}

void serialOutTicker_callback()
{

  static uint32_t SerialPrintCount = 0;

  if (SerialPrintCount % 10 == 0)
  {
    Serial.printf("| %8s | %8s | %8s |\n", FACTION_1_NAME, FACTION_2_NAME, FACTION_3_NAME);
  }
  Serial.printf("  %8d   %8d   %8d\n", Count_Faction_1, Count_Faction_2, Count_Faction_3);
  SerialPrintCount++;
}

void inputGetterTicker_callback()
{
  activeFaction = NONE;

  if (digitalRead(DIO_FAC_1_TRG) + digitalRead(DIO_FAC_2_TRG) + !digitalRead(DIO_FAC_3_TRG) < 2)
  {
    Serial.println("ERROR: More than one Flag active");
    return;
  }

  if (digitalRead(DIO_FAC_1_TRG) == LOW)
    activeFaction = FACTION_1;

  if (digitalRead(DIO_FAC_2_TRG) == LOW)
    activeFaction = FACTION_2;

  if (digitalRead(DIO_FAC_3_TRG) == HIGH)
    activeFaction = FACTION_3;
}

void setup()
{
  pinMode(DIO_FAC_1_TRG, INPUT_PULLUP);
  pinMode(DIO_FAC_2_TRG, INPUT_PULLUP);
  pinMode(DIO_FAC_3_TRG, INPUT);

  Serial.begin(9600);
  Serial.print("\n\n\n");

  LittleFS.begin();

  WiFi.persistent(false);
  WiFi.mode(WIFI_AP);
  WiFi.softAP(WIFI_SSID, WIFI_PASS);

  stationConnectedHandler = WiFi.onSoftAPModeStationConnected(&onStationConnected);
  stationDisconnectedHandler = WiFi.onSoftAPModeStationDisconnected(&onStationDisconnected);

#ifdef CAPTIVE
  dnsServer.start(53, "*", WiFi.softAPIP());
#endif
  server.addHandler(new CaptiveRequestHandler()).setFilter(ON_AP_FILTER); //only when requested from AP
  server.begin();

  FactionTicker.start();
  SerialOutputTicker.start();
  InputGetterTicker.start();
}

void loop()
{
  FactionTicker.update();
  SerialOutputTicker.update();
  InputGetterTicker.update();

#ifdef CAPTIVE
  dnsServer.processNextRequest();
#endif

  SevenSeg_Output();

  while (Serial.available() > 0)
  {
    char input = Serial.read();

    switch (input)
    {
    case 'n':
      activeFaction = NONE;
      break;

    case 'g':
      activeFaction = FACTION_1;
      break;

    case 'k':
      activeFaction = FACTION_3;
      break;

    case 'm':
      activeFaction = FACTION_2;
      break;

    case 'x':
      SerialOutputTicker.stop();
      Serial.println("SerialOutputTicker.stop()");
      break;

    case 'y':
      SerialOutputTicker.resume();
      Serial.println("SerialOutputTicker.resume()");
      break;

    default:
      break;
    }
  }
}